/*
Parrot
Copyright (C) 2003 Douglas Thain
http://www.cs.wisc.edu/~thain/research/parrot
This program is released under the GNU General Public License.
See the file COPYING for details.
*/


#include "chirp_protocol.h"

#include "debug.h"
#include "link.h"
#include "getopt.h"
#include "auth_all.h"
#include "stringtools.h"
#include "full_io.h"
#include "get_canonical_path.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <limits.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/statfs.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <signal.h>
#include <utime.h>

static void chirp_receive( struct link *l );
static void chirp_handler( struct link *l );
static int errno_to_chirp( int e );

static int port = CHIRP_PORT;
static int idle_timeout = 86400; /* one day */
static int stall_timeout = 3600; /* one hour */
static int single_mode = 0;
static const char *authfile = "/etc/chirp.authfile";

static void show_version( const char *cmd )
{
        printf("%s version %d.%d.%d built by %s@%s on %s at %s\n",cmd,DTTOOLS_VERSION_MAJOR,DTTOOLS_VERSION_MINOR,DTTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__DATE__,__TIME__);
}

static void show_help( const char *cmd )
{
	printf("use: %s [options]\n",cmd);
	printf("where options are:\n");
	printf(" -a <file>  Use this authorization file.\n");
	printf(" -d <flag>  Enable debugging for this sybsystem\n");
	printf(" -c <dir>   Challenge directory for filesystem authentication.\n");
	printf(" -h         This message.\n");
	printf(" -i         Use when the server is run under inetd.\n");
	printf(" -o <file>  Send debugging output to this file.\n");
	printf(" -O <bytes> Rotate file once it reaches this size.\n");
	printf(" -p <port>  Listen on this port (default is %d)\n",port);
	printf(" -s <time>  Abort stalled operations after this long. (default is 1h)\n");
	printf(" -S         Single process mode, do not fork.\n");
	printf(" -t <time>  Disconnect idle clients after this time. (default is 1d)\n");
	printf(" -v         Show version info.\n");

	printf("\n");
	printf("Where debug flags are: ");
	debug_flags_print(stdout);
	printf("\n\n");
	printf("Time arguments to -t and -s are assumed to be in seconds unless followed\n");
	printf("by an 's' 'm' 'h' or 'd' to indicate seconds, minutes, hours, or days.\n");
	printf("For example: chirp_server -t 7d -s 2h\n");
}

static int parse_time_string( const char *str )
{
	int value;
	char mod;

	if(sscanf(str,"%d%c",&value,&mod)==2) {
		switch(mod) {
			case 's': return value;
			case 'm': return value*60;
			case 'h': return value*60*60;
			case 'd': return value*60*60*24;
		}
	} else if(sscanf(str,"%d",&value)==1) {
		return value;
	}

	fprintf(stderr,"Sorry, I don't understand the time value '%s'\n",str);
	exit(1);
}

static void install_handler( int sig, void (*handler)(int sig))
{
	struct sigaction s;

	s.sa_handler = handler;
	sigfillset(&s.sa_mask);
	s.sa_flags = 0; 

	sigaction(sig,&s,0);
}

void ignore_signal( int sig )
{
}

int main( int argc, char *argv[] )
{
	struct link *link;
	int inetd_mode = 0;
	char c;

	debug_config(argv[0]);

	install_handler(SIGCHLD,ignore_signal);

	while((c=getopt(argc,argv,"a:c:t:s:Sp:o:O:d:vhi"))!=(char)-1) {
		switch(c) {
			case 'a':
				authfile = optarg;
				break;
			case 'c':
				auth_filesystem_challenge_dir(optarg);
				break;
			case 'p':
				port = atoi(optarg);
				break;
			case 't':
				idle_timeout = parse_time_string(optarg);
				break;
			case 'i':
				inetd_mode = 1;
				break;
			case 's':
				stall_timeout = parse_time_string(optarg);
				break;
			case 'S':
				single_mode = 1;
				break;
			case 'd':
				debug_flags_set(optarg);
				break;
			case 'o':
				debug_config_file(optarg);
				break;
			case 'O':
				debug_config_file_size(atoi(optarg));
				break;
			case 'v':
				show_version(argv[0]);
				return 1;
			case 'h':
			default:
				show_help(argv[0]);
				return 1;
		}
	}

	if(access(authfile,R_OK)!=0) fatal("couldn't read authorization file %s: %s",authfile,strerror(errno));
		
	auth_register_all();

	if(inetd_mode) {
		struct link *l = link_attach(0);
		if(l) {
			chirp_receive(l);
			exit(0);
		} else {
			exit(1);
		}
	}

	link = link_serve(port);
	if(!link) fatal("couldn't listen on port %d: %s",port,strerror(errno));

	while(1) {
		struct link *l;
		pid_t pid;

		l = link_accept(link,time(0)+stall_timeout);
		if(!l) continue;

		if(single_mode) {
			chirp_receive(l);
		} else {
			pid = fork();
			if(pid==0) {
				chirp_receive(l);
				_exit(0);
			} else if(pid>=0) {
				debug(D_CHIRP,"created pid %d",pid);
			} else {
				debug(D_CHIRP,"couldn't fork: %s",strerror(errno));
			}
			link_close(l);
		}
	}
}

static void chirp_receive( struct link *link )
{
	struct passwd *p;
	char *atype, *ahost, *asubject, *username;
	char addr[LINK_ADDRESS_MAX];
	int port;

	link_address_remote(link,addr,&port);

	if(auth_accept(link,authfile,&atype,&ahost,&asubject,&username,time(0)+stall_timeout)) {
		p = getpwnam(username);
		if(p) {
			debug(D_AUTH,"%s is %s uid %d",addr,username,p->pw_uid);
			if(getuid()==0) {
				debug(D_AUTH,"changing to uid %d",p->pw_uid);
				if(setuid(p->pw_uid)==0) {
					chirp_handler(link);
				} else {
					debug(D_AUTH,"setuid failed: %s",strerror(errno));
				}
			} else if(getuid()==p->pw_uid) {
				chirp_handler(link);
			} else {
				debug(D_AUTH,"sorry, but I am not %s",username);
			}
		} else {
			debug(D_AUTH,"couldn't identify username %s",username);
		}
		free(atype);
		free(ahost);
		free(asubject);
		free(username);
	} else {
		debug(D_AUTH,"client could not authenticate");
	}

	link_close(link);
}

static void chirp_handler( struct link *l )
{
	char line[CHIRP_LINE_MAX];
	
	link_tune(l,LINK_TUNE_INTERACTIVE);

	while(1) {
		int do_stat_result = 0;
		int do_statfs_result = 0;
		int do_getdir_result = 0;
		int do_no_result = 0;

		char *dataout = 0;
		int dataoutlength = 0;

		char path[CHIRP_PATH_MAX];
		char newpath[CHIRP_PATH_MAX];
		int fd, offset, length, flags;
		int result, actual;
		uid_t uid;
		gid_t gid;
		mode_t mode;
		struct stat statbuf;
		struct statfs statfsbuf;
		struct utimbuf ut;
		time_t idletime = time(0)+idle_timeout;
		time_t stalltime;

		if(!link_readline(l,line,sizeof(line),idletime)) {
			if(time(0)>=idletime) {
				debug(D_CHIRP,"timeout: client idle too long\n");
				sprintf(line,"%d\n",CHIRP_ERROR_TIMED_OUT);
				link_write(l,line,strlen(line),time(0)+stall_timeout);
			}
			break;
		}

		stalltime = time(0)+stall_timeout;
		string_chomp(line);
		if(strlen(line)<1) continue;
		if(line[0]==4) break;

		debug(D_CHIRP,"%s",line);

		if(sscanf(line,"pread %d %d %d",&fd,&length,&offset)==3) {
			dataout = malloc(length);
			if(dataout) {
				result = full_pread(fd,dataout,length,offset);
				if(result>=0) {
					dataoutlength = result;
				} else {
					free(dataout);
					dataout = 0;
				}
			} else {
				errno = ENOMEM;
			}
		} else if(sscanf(line,"readlink %s %d",path,&length)==2) {
			dataout = malloc(length);
			if(dataout) {
				result = readlink(path,dataout,length);
				if(result>=0) {
					dataoutlength = result;
				} else {
					free(dataout);
					dataout = 0;
				}
			} else {
				errno = ENOMEM;
			}
		} else if(sscanf(line,"canonicalize %s %d",path,&length)==2) {
			dataout = malloc(length);
			if(dataout) {
				result = get_canonical_path(path,dataout,length);
				if(result>=0) {
					dataoutlength = result;
				} else {
					free(dataout);
					dataout = 0;
				}
			} else {
				errno = ENOMEM;
			}
		} else if(sscanf(line,"pwrite %d %d %d",&fd,&length,&offset)==3) {
			char *data = malloc(length);
			if(data) {
				actual = link_read(l,data,length,stalltime);
				if(actual!=length) break;

				result = full_pwrite(fd,data,length,offset);
				free(data);
			} else {
				/* we ought to clear the stream and continue here */
				break;
			}
		} else if(sscanf(line,"getdir %s",path)==1) {
			DIR *dir;
			struct dirent *d;

			dir = opendir(path);
			if(dir) {
				link_write(l,"0\n",2,stalltime);
				while((d=readdir(dir))) {
					sprintf(line,"%s\n",d->d_name);
					link_write(l,line,strlen(line),stalltime);
				}
				closedir(dir);
				do_getdir_result = 1;
				result = 0;
			} else {
				result = -1;
			}

		} else if(sscanf(line,"getfile %s",path)) {
			fd = open(path,O_RDONLY);
			if(fd>=0) {
				length = lseek(fd,0,SEEK_END);
				lseek(fd,0,SEEK_SET);
				sprintf(line,"%d\n",length);
				link_write(l,line,strlen(line),stalltime);
				result = link_stream_from_fd(l,fd,length,stalltime);
				close(fd);
				if(result!=length) break;
				do_no_result = 1;
			} else {
				result = -1;
			}

		} else if(sscanf(line,"putfile %s %d %d",path,&mode,&length)) {
			fd = open(path,O_WRONLY|O_CREAT|O_TRUNC,mode);
			if(fd>=0) {
				actual = link_stream_to_fd(l,fd,length,stalltime);
				if(actual!=length) {
					if(actual>=0) link_soak(l,length-actual,stalltime);
					result = -1;
				} else {
					result = actual;
				}
				close(fd);
			} else {
				result = -1;
			}

		} else if(sscanf(line,"open %s %s %d",path,newpath,&mode)==3) {
			flags = 0;

			if(strchr(newpath,'r')) {
				if(strchr(newpath,'w')) {
					flags = O_RDWR;
				} else {
					flags = O_RDONLY;
				}
			} else if(strchr(newpath,'w')) {
				flags = O_WRONLY;
			}

			if(strchr(newpath,'c')) flags |= O_CREAT;
			if(strchr(newpath,'t')) flags |= O_TRUNC;
			if(strchr(newpath,'a')) flags |= O_APPEND;
			if(strchr(newpath,'x')) flags |= O_EXCL;
			if(strchr(newpath,'s')) flags |= O_SYNC;

			result = open(path,flags,mode);
			if(result>=0) {
				fstat(result,&statbuf);
				do_stat_result = 1;
			}
		} else if(sscanf(line,"close %d",&fd)==1) {
			result = close(fd);
		} else if(sscanf(line,"fchmod %d %d",&fd,&mode)==2) {
			result = fchmod(fd,mode);
		} else if(sscanf(line,"fchown %d %d %d",&fd,&uid,&gid)==3) {
			result = fchown(fd,uid,gid);
		} else if(sscanf(line,"fsync %d",&fd)==1) {
			result = fsync(fd);
		} else if(sscanf(line,"ftruncate %d %d",&fd,&length)==2) {
			result = ftruncate(fd,length);
		} else if(sscanf(line,"unlink %s",path)==1) {
			result = unlink(path);
		} else if(sscanf(line,"access %s %d",path,&mode)==2) {
			result = access(path,mode);
		} else if(sscanf(line,"chmod %s %d",path,&mode)==2) {
			result = chmod(path,mode);
		} else if(sscanf(line,"chown %s %d %d",path,&uid,&gid)==3) {
			result = chown(path,uid,gid);
		} else if(sscanf(line,"lchown %s %d %d",path,&uid,&gid)==3) {
			result = lchown(path,uid,gid);
		} else if(sscanf(line,"truncate %s %d",path,&length)==2) {
			result = truncate(path,length);
		} else if(sscanf(line,"rename %s %s",path,newpath)==2) {
			result = rename(path,newpath);
		} else if(sscanf(line,"link %s %s",path,newpath)==2) {
			result = link(path,newpath);
		} else if(sscanf(line,"symlink %s %s",path,newpath)==2) {
			result = symlink(path,newpath);
		} else if(sscanf(line,"mkdir %s %d",path,&mode)==2) {
			result = mkdir(path,mode);
		} else if(sscanf(line,"rmdir %s",path)==1) {
			result = rmdir(path);
		} else if(sscanf(line,"utime %s %lu %lu",path,&ut.actime,&ut.modtime)==3) {
			result = utime(path,&ut);
		} else if(sscanf(line,"fstatfs %d",&fd)==1) {
			result = fstatfs(fd,&statfsbuf);
			do_statfs_result = 1;
		} else if(sscanf(line,"statfs %s",path)==1) {
			result = statfs(path,&statfsbuf);
			do_statfs_result = 1;
		} else if(sscanf(line,"fstat %d",&fd)==1) {
			result = fstat(fd,&statbuf);
			do_stat_result = 1;
		} else if(sscanf(line,"stat %s",path)==1) {
			result = stat(path,&statbuf);
			do_stat_result = 1;
		} else if(sscanf(line,"lstat %s",path)==1) {
			result = lstat(path,&statbuf);
			do_stat_result = 1;
		} else {
			result = -1;
			errno = ENOSYS;
		}

		if(do_no_result) {
			/* nothing */
		} else if(result<0) {
			result = errno_to_chirp(errno);
			sprintf(line,"%d\n",result);
		} else if(do_stat_result) {
			sprintf(line,"%lld\n%lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld\n",
				(long long) result,
				(long long) statbuf.st_dev,
				(long long) statbuf.st_ino,
				(long long) statbuf.st_mode,
				(long long) statbuf.st_nlink,
				(long long) statbuf.st_uid,
				(long long) statbuf.st_gid,
				(long long) statbuf.st_rdev,
				(long long) statbuf.st_size,
				(long long) statbuf.st_blksize,
				(long long) statbuf.st_blocks,
				(long long) statbuf.st_atime,
				(long long) statbuf.st_mtime,
				(long long) statbuf.st_ctime);
		} else if(do_statfs_result) {
			sprintf(line,"%lld\n%lld %lld %lld %lld %lld %lld %lld\n",
				(long long) result,
				(long long) statfsbuf.f_type,
       				(long long) statfsbuf.f_bsize,
				(long long) statfsbuf.f_blocks,
				(long long) statfsbuf.f_bfree,
				(long long) statfsbuf.f_bavail,
				(long long) statfsbuf.f_files,
				(long long) statfsbuf.f_ffree);
		} else if(do_getdir_result) {
			sprintf(line,"\n");
		} else {
			sprintf(line,"%d\n",result);
		}

		debug(D_CHIRP,"= %s",line);
		if(!do_no_result) {
			length = strlen(line);
			actual = link_write(l,line,length,stalltime);
			if(actual!=length) break;
		}

		if(dataout) {
			actual = link_write(l,dataout,dataoutlength,stalltime);
			if(actual!=dataoutlength) break;
			free(dataout);
			dataout = 0;
		}

	}
}

static int errno_to_chirp( int e )
{
	switch(e) {
		case EACCES:
		case EPERM:
			return CHIRP_ERROR_NOT_AUTHORIZED;
		case ENOENT:
			return CHIRP_ERROR_DOESNT_EXIST;
		case EEXIST:
			return CHIRP_ERROR_ALREADY_EXISTS;
		case EFBIG:
			return CHIRP_ERROR_TOO_BIG;
		case ENOSPC:
			return CHIRP_ERROR_NO_SPACE;
		case ENOMEM:
			return CHIRP_ERROR_NO_MEMORY;
		case ENOSYS:
		case EINVAL:
			return CHIRP_ERROR_INVALID_REQUEST;
		case EMFILE:
		case ENFILE:
			return CHIRP_ERROR_TOO_MANY_OPEN;
		case EBUSY:
			return CHIRP_ERROR_BUSY;
		case EAGAIN:
			return CHIRP_ERROR_TRY_AGAIN;
		case EBADF:
			return CHIRP_ERROR_BAD_FD;
		case EISDIR:
			return CHIRP_ERROR_IS_DIR;
		case ENOTDIR:
			return CHIRP_ERROR_NOT_DIR;
		case ENOTEMPTY:
			return CHIRP_ERROR_NOT_EMPTY;
		case EXDEV:
			return CHIRP_ERROR_CROSS_DEVICE_LINK;
		default:
			debug(D_CHIRP,"zoiks, I don't know how to transform error %d (%s)\n",errno,strerror(errno));
			return CHIRP_ERROR_UNKNOWN;
	}
}
