/*
PFS is Copyright (C) 2001 Douglas Thain.
This program is released under the GNU General Public License.
See the file COPYING for details.
*/

#include "pfs_service.h"

extern "C" {
#include "debug.h"
#include "get_canonical_path.h"
}

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/vfs.h>
#include <sys/file.h>
#include <sys/time.h>
#include <utime.h>
#include <dirent.h>
#include <sys/poll.h>

#define END debug(D_LOCAL,"= %d %s",result,((result>=0) ? "" : strerror(errno)) ); return result;

extern "C" ssize_t pread(int fd, void *buf, size_t count, off_t  offset);
extern "C" ssize_t pwrite(int  fd,  const  void  *buf, size_t count, off_t offset);

class pfs_file_local : public pfs_file
{
private:
	int fd;
	pfs_off_t last_offset;

public:
	pfs_file_local( pfs_name *name, int f ) : pfs_file(name) {
		fd = f;
		last_offset = 0;
	}

	virtual int close() {
		int result;
		debug(D_LOCAL,"close %d",fd);
		result = ::close(fd);
		END
	}

	virtual pfs_ssize_t read( void *data, pfs_size_t length, pfs_off_t offset ) {
		pfs_ssize_t result;
		debug(D_LOCAL,"read %d 0x%x %d %d",fd,data,length,offset);
		if(offset!=last_offset) ::lseek(fd,offset,SEEK_SET);
		result = ::read(fd,data,length);
		if(result>0) last_offset = offset+result;
		END
	}

	virtual pfs_ssize_t write( const void *data, pfs_size_t length, pfs_off_t offset ) {
		pfs_ssize_t result;
		debug(D_LOCAL,"write %d 0x%x %d %d",fd,data,length,offset);
		if(offset!=last_offset) ::lseek(fd,offset,SEEK_SET);
		result = ::write(fd,data,length);
		if(result>0) last_offset = offset+result;
		END
	}

	virtual int fstat( struct stat *buf ) {
		int result;
		debug(D_LOCAL,"fstat %d 0x%x",fd,buf);
		result = ::fstat(fd,buf);
		END
	}

	virtual int fstatfs( struct statfs *buf ) {
		int result;
		debug(D_LOCAL,"fstatfs %d 0x%x",fd,buf);
		result = ::fstatfs(fd,buf);
		END
	}

	virtual int ftruncate( size_t length ) {
		int result;
		debug(D_LOCAL,"truncate %d %d",fd,length);
		result = ::ftruncate(fd,length);
		END
	}

	virtual int fsync() {
		int result;
		debug(D_LOCAL,"fsync %d",fd);
		result = ::fsync(fd);
		END
	}

	virtual int fcntl( int cmd, void *arg ) {
		int result;
		debug(D_LOCAL,"fcntl %d %d 0x%x",fd,cmd,arg);
		if(cmd==F_SETFL) arg = (void*)(((int)arg)|O_NONBLOCK);
		result = ::fcntl(fd,cmd,arg);
		END
	}

	virtual int ioctl( int cmd, void *arg ) {
		int result;
		debug(D_LOCAL,"ioctl %d 0x%x 0x%x",fd,cmd,arg);
		result = ::ioctl(fd,cmd,arg);
		END
	}

	virtual int fchmod( mode_t mode ) {
		int result;
		debug(D_LOCAL,"fchmod %d %d",fd,mode);
		result = ::fchmod(fd,mode);
		END
	}

	virtual int fchown( uid_t uid, gid_t gid ) {
		int result;
		debug(D_LOCAL,"fchown %d %d %d",fd,uid,gid);
		result = ::fchown(fd,uid,gid);
		END
	}

	virtual int flock( int op ) {
		int result;
		debug(D_LOCAL,"flock %d %d",fd,op);
		result = ::flock(fd,op);
		END
	}

	virtual void * mmap( void *start, pfs_size_t length, int prot, int flags, off_t offset ) {
		void *result;
		result = ::mmap((caddr_t)start,length,prot,flags,fd,offset);
		END
	}

	virtual pfs_ssize_t get_size() {
		struct stat s;
		int result;
		result = this->fstat(&s);
		if(result<0) {
			return 0;
		} else {
			return s.st_size;
		}
	}
	virtual int get_real_fd() {
		return fd;
	}

	virtual int get_local_name( char *n )
	{
		strcpy(n,name.rest);
		return 0;
	}

	virtual void poll_register( int which ) {
		pfs_poll_wakeon(fd,which);
	}

	virtual int poll_ready() {
		struct pollfd pfd;
		int result=0, flags =0;

		pfd.fd = fd;
		pfd.events = POLLIN|POLLOUT|POLLERR;
		pfd.revents = 0;

		result = ::poll(&pfd,1,0);
		if(result>0) {
			if(pfd.revents&POLLIN) flags |= PFS_POLL_READ;
			if(pfd.revents&POLLOUT) flags |= PFS_POLL_WRITE;
			if(pfd.revents&POLLERR) flags |= PFS_POLL_EXCEPT;
		}
		return flags;
	}
};

class pfs_service_local : public pfs_service {
public:
	virtual pfs_file * open( pfs_name *name, int flags, mode_t mode ) {
		pfs_file *result;
		flags |= O_NONBLOCK;
		debug(D_LOCAL,"open %s %d %d",name->rest,flags,(flags&O_CREAT) ? mode : 0);
		int fd = ::open(name->rest,flags,mode);
		if(fd>=0) {
			result = new pfs_file_local(name,fd);
		} else {
			result = 0;
		}
		debug(D_LOCAL,"= %d %s",fd,(fd>=0) ? "" : strerror(errno));
		return result;
	}

	virtual pfs_dir * getdir( pfs_name *name ) {
		struct dirent *d;
		DIR *dir;
		pfs_dir *result = 0;
		debug(D_LOCAL,"getdir %s",name->rest);
		dir = ::opendir(name->rest);
		if(dir) {
			result = new pfs_dir(name);
			while((d=::readdir(dir))) {
				result->append(d->d_name);
			}
			closedir(dir);
		} else {
			result = 0;
       		}
		debug(D_LOCAL,"= %s",result ? "Success" : strerror(errno));
		return result;
	}

	virtual int stat( pfs_name *name, struct stat *buf ) {
		int result;
		debug(D_LOCAL,"stat %s 0x%x",name->rest,buf);
		result = ::stat(name->rest,buf);
		END
	}
	virtual int statfs( pfs_name *name, struct statfs *buf ) {
		int result;
		debug(D_LOCAL,"statfs %s 0x%x",name->rest,buf);
		result = ::statfs(name->rest,buf);
		END
	}
	virtual int lstat( pfs_name *name, struct stat *buf ) {
		int result;
		debug(D_LOCAL,"lstat %s 0x%x",name->rest,buf);
		result = ::lstat(name->rest,buf);
		END
	}
	virtual int access( pfs_name *name, mode_t mode ) {
		int result;
		debug(D_LOCAL,"access %s %d",name->rest,mode);
		result = ::access(name->rest,mode);
		END
	}
	virtual int chmod( pfs_name *name, mode_t mode ) {
		int result;
		debug(D_LOCAL,"chmod %s %d",name->rest,mode);
		result = ::chmod(name->rest,mode);
		END
	}
	virtual int chown( pfs_name *name, uid_t uid, gid_t gid ) {
		int result;
		debug(D_LOCAL,"chown %s %d %d",name->rest,uid,gid);
		result = ::chown(name->rest,uid,gid);
		END
	}
	virtual int lchown( pfs_name *name, uid_t uid, gid_t gid ) {
		int result;
		debug(D_LOCAL,"lchown %s %d %d",name->rest,uid,gid);
		result = ::lchown(name->rest,uid,gid);
		END
	}
	virtual int truncate( pfs_name *name, pfs_off_t length ) {
		int result;
		debug(D_LOCAL,"truncate %s %d",name->rest,length);
		result = ::truncate(name->rest,length);
		END
	}
	virtual int utime( pfs_name *name, struct utimbuf *buf ) {
		int result;
		debug(D_LOCAL,"utime %s 0x%x",name->rest,buf);
		result = ::utime(name->rest,buf);
		END
	}
	virtual int unlink( pfs_name *name ) {
		int result;
		debug(D_LOCAL,"unlink %s",name->rest);
		result = ::unlink(name->rest);
		END
	}
	virtual int rename( pfs_name *oldname, pfs_name *newname ) {
		int result;
		debug(D_LOCAL,"rename %s %s",oldname->rest,newname->rest);
		result = ::rename(oldname->rest,newname->rest);
		END
	}

	/*
	We do not actually change to the new directory,
	because this is performed within the PFS master
	process, and we do not want to change the meaning
	of open() on filenames used for configuration,
	security and so forth.  We also do not change and then
	move back, because what will we do if the chdir
	back fails?
	*/
	virtual int chdir( pfs_name *name, char *newpath ) {
		int result;
		debug(D_LOCAL,"canonicalize %s",name->rest);
		result = ::get_canonical_path(name->rest,newpath,PFS_PATH_MAX);
		END
	}

	virtual int link( pfs_name *oldname, pfs_name *newname ) {
		int result;
		debug(D_LOCAL,"link %s %s",oldname->rest,newname->rest);
		result = ::link(oldname->rest,newname->rest);
		END
	}
	virtual int symlink( pfs_name *oldname, pfs_name *newname ) {
		int result;
		debug(D_LOCAL,"symlink %s %s",oldname->rest,newname->rest);
		result = ::symlink(oldname->rest,newname->rest);
		END
	}
	virtual int readlink( pfs_name *name, char *buf, pfs_size_t size ) {
		int result;
		debug(D_LOCAL,"readlink %s 0x%x %d",name->rest,buf,size);
		result = ::readlink(name->rest,buf,size);
		END
	}
	virtual int mkdir( pfs_name *name, mode_t mode ) {
		int result;
		debug(D_LOCAL,"mkdir %s %d",name->rest,mode);
		result = ::mkdir(name->rest,mode);
		END
	}
	virtual int rmdir( pfs_name *name ) {
		int result;
		debug(D_LOCAL,"rmdir %s",name->rest);
		result = ::rmdir(name->rest);
		END
	}

	virtual int is_seekable() {
		return 1;
	}

	virtual int is_local() {
		return 1;
	}
};

static pfs_service_local pfs_service_local_instance;
pfs_service *pfs_service_local = &pfs_service_local_instance;

struct pfs_file * pfs_file_bootstrap( int fd, const char *path )
{
	pfs_name name;
	name.service = pfs_service_local;
	strcpy(name.path,path);
	strcpy(name.service_name,"local");
	name.host[0] = 0;
	name.port = 0;
	strcpy(name.rest,path);
	name.is_local = 1;
	return new pfs_file_local(&name,fd);
}
